#define _c_ASCIINPE
/*********************************************************************/
/*                                                                   */
/* MODULE NAME: ASCIINPE.C                                           */
/*                                                                   */
/*                                                                   */
/* SYNOPSIS:  ACIF Sample ASCII datastream input translation exit    */
/*            plus ASCII to EBCDIC conversion.                       */
/*                                                                   */
/*                                                                   */
/* DESCRIPTION:  This module contains a sample exit that converts    */
/*               ASCII carriage return and formfeed characters       */
/*               and performs ASCII to EBCDIC conversion.            */
/*                                                                   */
/* COPYRIGHT:                                                        */
/*  5765-140,5648-062 (C) COPYRIGHT IBM CORPORATION 1987,1993.       */
/*  All Rights Reserved                                              */
/*  Licensed Materials - Property of IBM                             */
/*                                                                   */
/*  US Government Users Restricted Rights - Use, duplication or      */
/*  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.*/
/*                                                                   */
/* NOTE: This program sample is provided on an as-is basis.          */
/*       The licensee of the PSF product is free to copy, revise,    */
/*       modify, and make derivative works of this program sample    */
/*       as they see fit.                                            */
/*                                                                   */
/* STATUS = RELEASE 1  LEVEL 0                                       */
/*                                                                   */
/*                                                                   */
/* EXTERNAL STATIC VARIABLES: None                                   */
/*                                                                   */
/*                                                                   */
/* NOTES:                                                            */
/*                                                                   */
/*                                                                   */
/* CHANGE ACTIVITY:                                                  */
/*                                                                   */
/* PROBNO  REL   DATE/AUTHOR DESCRIPTION                        FICHE*/
/* ------- ----- ----------- -----------                        -----*/
/*               7/94 T. Medearis  See below for description         */
/*                                                                   */
/*********************************************************************/
/*  This program is a sample exit for ACIF (APKACIF,5648-062)        */
/*  that accepts ASCII datastreams containing carriage return        */
/*  and Formfeed characters and allows them to be formatted as       */
/*  linedata using a PAGEDEF.  It then translates the input          */
/*  records from ASCII to EBCDIC.                                    */
/*                                                                   */
/*                                                                   */
/*  You must supply the ACIF control statement INPEXIT=ASCIINPE      */
/*  to enable this exit, and include the load module in a            */
/*  dataset referenced in STEPLIB or elsewhere in the module         */
/*  search order.  You must also specify the control statments       */
/*  CC=YES and CCTYPE=Z.                                             */
/*                                                                   */
/*                                                                   */
/*  Method of Operation:                                             */
/*                                                                   */
/*  The basic idea behind this exit is to transform an ASCII         */
/*  datastream into a record format containing a carriage control    */
/*  character in byte 0 of every record and then translate the       */
/*  record from ASCII to EBCDIC.                                     */
/*                                                                   */
/*  For each input record, a carriage control character is inserted  */
/*  in front of the record in byte 0 and the length of the record is */
/*  adjusted.  However, if the byte 0 of the input record is an      */
/*  ASCII carriage return (x'0C') or an ASCII formfeed character     */
/*  (x'0D'), then no character is inserted.  Instead byte 0 is       */
/*  changed to an ASCII space (x'20'), which in the carriage control */
/*  byte serves to return the carriage and advance one line, or to   */
/*  an ANSI skip to channel 1 command (x'31') which in the carriage  */
/*  control byte serves as a formfeed.                               */
/*                                                                   */
/*  Processing then continues to search for the next carriage return */
/*  (x'0C') or formfeed (x'0D') character.  When such a character is */
/*  found, any remaining data from the input record, including the   */
/*  unchanged carriage return or formfeed character is placed into   */
/*  storage, the ACIF REQUEST byte is set to insert (x'02' - this    */
/*  causes ACIF to invoke the exit again without sending a new input */
/*  record, so that the exit may process the data that was placed in */
/*  storage on the previous call), the data up to but not including  */
/*  the carriage return or formfeed character is returned to ACIF,   */
/*  and control is returned to ACIF.                                 */
/*                                                                   */
/*  If the end of the record is reached and no carriage return or    */
/*  formfeed character is found, the data will be returned with the  */
/*  REQUEST byte set to use (x'00') indicating that all data has     */
/*  been processed by the exit, and ACIF may invoke the exit with a  */
/*  new input record.  If the last byte of a record from storage or  */
/*  from initial input is a carriage return character, the byte is   */
/*  deleted by adjusting the length of the data returned to ACIF.    */
/*                                                                   */
/*  If control returns to ACIF with the REQUEST byte set to INSERT,  */
/*  ACIF will then call the exit again, this time with no new data,  */
/*  because data remains in storage that still needs to be processed */
/*  by the exit.  Since the data is from storage, the first byte     */
/*  should be a carriage return or formfeed character.  This byte    */
/*  will be translated as described above for formfeed characters.   */
/*  However, carriage return characters will be translated into      */
/*  plus (x'2B') carriage control characters which cause the carriage*/
/*  to return but do not force a line feed.  This is done because it */
/*  is assumed that the carriage return was meant to allow print     */
/*  on the same line, such as underscore characters.  Processing then*/
/*  continues to search for the next carriage return or formfeed     */
/*  character as described above until all the records from the      */
/*  ASCII datastream have been processed.                            */
/*                                                                   */
/*  Each record is then translated from ASCII to EBCDIC.             */
/*                                                                   */
/*  Thus, carriage return characters (x'0C') are translated into     */
/*  either space (x'20') or plus (x'2B') carriage control characters,*/
/*  and formfeed characters (x'0D') are translated into skip to      */
/*  channel 1 (x'31') carriage control characters.                   */
/*                                                                   */
/*  No other ASCII printer commands are recognized, but this         */
/*  exit could easily be modified to account for vertical tab        */
/*  (x'0B'), horizontal tab (x'09'), and backspace (x'08').          */
/*  See 'EXTENDING THIS EXIT' below for details.                     */
/*                                                                   */
/* Pseudocode:                                                       */
/*   Initialize variables                                            */
/*   Scan for ASCII Carriage Return or Formfeed characters           */
/*     If (CR or FF is found in first byte)                          */
/*         Translate character and continue searching                */
/*     If (CR or FF is found in middle or last byte)                 */
/*         Set flag to discontinue scan.                             */
/*   If data remains to be examined                                  */
/*      if storage has not already been obtained                     */
/*         get storage for remaining data                            */
/*         move remaining data from input area to storage area       */
/*      else                                                         */
/*         update storage record pointer                             */
/*   If working with data in input area, shift data if necessary     */
/*      to make room for space that needs to be inserted.            */
/*   CALL ASC2EB to convert the record from ASCII to EBCDIC.         */
/*   If all data has been examined, free any storage that may have   */
/*      been obtained.                                               */
/*                                                                   */
/* EXTENDING THIS EXIT:                                              */
/*                                                                   */
/*   The key to extending this exit is the SWITCH statement in the   */
/*   WHILE loop.                                                     */
/*   If it is necessary to convert additional input characters,      */
/*    1. define a constant for the input character                   */
/*    2. define a constant for the desired output character          */
/*    3. add a CASE section within the WHILE loop for the input      */
/*       char.  The following fields will need to be set             */
/*       accordingly.                                                */
/*       a. exitstruc->request - set depending upon whether or not   */
/*            additional data remains to be examined.  If it is set  */
/*            to INP_USE, then all the data has been examined, and   */
/*            any obtained storage should be freed.  If it is set to */
/*            INP_INSERT, then data remains to be examined in        */
/*            storage.                                               */
/*       b. DONE - flag to exit the loop.  For Formfeed and          */
/*            Carriage Return, DONE is set to true whenever a FF or  */
/*            CR is found after the first byte in the data being     */
/*            examined.  If found in the first byte, the byte is     */
/*            translated accordingly and the search continues.       */
/*       c. switches.shiftinp - For the FF and CR cases, this flag   */
/*            is set to false if a FF or CR is found in the first    */
/*            byte.  This flag change is only really needed if the   */
/*            data being examined is from an initial call.  The      */
/*            initial setting of this flag based on whether or not   */
/*            the call is examining data from storage or from an     */
/*            initial call.  It is always set to false if examining  */
/*            data in storage.  If examining data from input, then   */
/*            a carriage return and line feed character is inserted  */
/*            in front of the record.  All subsequent instances of   */
/*            a carriage return or formfeed character in the data    */
/*            are translated to ascii carriage control characters.   */
/*            Subsequent carriage return charcters do not result in  */
/*            a line feed, because it is assumed that the intent is  */
/*            do perform operations on the same line such as under-  */
/*            scoring.
/*       d. exitstruc[recindex] - set this field as desired in order */
/*            to make the translation.  Note that translation is     */
/*            only done when the found char is in the first byte     */
/*            of the data being scanned.  Otherwise, the char is     */
/*            left alone and the DONE flag is set to true.  This     */
/*            allows the record up to that point but not including   */
/*            the translate character to be returned in a record.    */
/*            the translate character will then be found in the      */
/*            first byte of data in storage when the exit is next    */
/*            called, since the remaining data is placed in storage  */
/*            and the exitstruc->request flage is set to INSERT      */
/*       e. switches.delchar - if a character is to be deleted, such */
/*            as a trailing carriage return char, then set this      */
/*            flag to true.  Currently, a byte may be deleted only   */
/*            if it is the last byte of the record.                  */
/*       f. switches.moredata - if a character is found in the       */
/*            middle of a record, then data remains to be examined   */
/*            and translated.  Set moredata to true.                 */
/*                                                                   */
/*********************************************************************/
#include "apkexits.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

long ASC2EB( INPEXIT_PARMS *exitstruc );

#define MAX_LEN 32756          /* definition of length of storage area*/
#define false  0
#define true 1
#define first  0               /* index of first byte of a record     */
#define skip2ch1 0x31          /* skip to channel 1 constant          */
#define space    0x20          /* space - print the line and then     */
                               /* carriage return and advance one line*/
#define formfeed 0x0C          /* formfeed constant                   */
#define crgret   0x0D          /* carriage return                     */
#define prntline 0x2B          /* print the line, do not carriage     */
                               /* return, do not linefeed.            */

struct wrkarea
 {
   char    *stgbasptr;         /* pointer to 32k storage area in      */
                               /* which portion of record that        */
                               /* must be stored across multiple      */
                               /* calls is stored.                    */
   char    *stgrecptr;         /* pointer to data to be examined in   */
                               /* the 32K storage area                */
   int     stgreclen;          /* length of data pointed to by        */
                               /* stgrecptr.                          */
   char    res[4];             /* reserved                            */
 };

struct on_off
 {
   unsigned resrvd   : 28;
   unsigned moredata : 1;    /* more data remains to be examined.     */
   unsigned fromstg  : 1;    /* working with data from storage        */
   unsigned shiftinp : 1;    /* data needs to be shifted one byte     */
   unsigned delchar  : 1;    /* delete the last char of the data to   */
                             /* be returned in exitstruc              */
 } switches;

struct wrkarea *workarea;    /* pointer to the 16 byte work area      */
int    remlength;            /* remaining length to be copied to the  */
                             /* storage and used on the next invoca-  */
                             /* tion of the exit.                     */
int    recindex;             /* index into the input record.          */
int    DONE;                 /* flag that indicates completion of loop*/
                             /* processing                            */
int    num=1;                /* number of arrays of data to allocate  */
                             /*  on calloc invocation.                */
char   *storTO;              /* pointer to where data is to be moved  */
char   *inputTO;             /* pointer to where data is to be moved  */
char   *FROM;                /* pointer to where data is to be moved  */
                             /*   from.                               */
char   *recordptr;           /* pointer to record to be analyzed      */
int    recordlen;            /* length of record to be analyzed       */
int    returnlen;            /* length of record to be returned       */
int    MOVESTG;              /* should data be moved from storage     */
int    maxindex;             /* max byte offset into analyzed record  */
long   rc;                   /* return code from ASC2EB               */

long INPEXIT( INPEXIT_PARMS *exitstruc)
{

     /***********************************************************/
     /*  Initialize ALL local variables.                        */
     /***********************************************************/
   switches.delchar = false;
   switches.moredata = false;
   rc = 0;
   remlength = 0;
   recindex = 0;
   recordptr = 0;
   recordlen = 0;
   returnlen = 0;
   maxindex = 0;
   exitstruc->request = INP_USE; /* set no call back needed default  */

   workarea = (struct wrkarea *) exitstruc->work;
                                    /* set pointer to workarea          */

   DONE = false;                    /* initialize DONE to false         */


     /***********************************************************/
     /*  If data from a previous call needs to be processed,    */
     /*  then set the record pointer to point to the data in    */
     /*  storage, and obtain the length of the record.          */
     /***********************************************************/

   if (workarea->stgreclen != 0)
    {
      recordptr = workarea->stgrecptr; /* work with record in storage   */
      recordlen = workarea->stgreclen;
      switches.fromstg = true;           /* set flag indicating working  */
                                         /* with data left over from a   */
                                         /* previous call.               */
      switches.shiftinp = false;         /* shift is never necessary when*/
                                         /* working with data in storage */
    }
   else
    {
      recordptr = exitstruc->record;     /* work with record from input  */
      recordlen = exitstruc->recordln;
      switches.fromstg = false;
      switches.shiftinp = true;          /* set default for shift to true*/

     }

   maxindex = recordlen;                 /* max index for while loop     */


     /***********************************************************/
     /*  Search the record for a formfeed or a carriage control.*/
     /***********************************************************/

   while ((DONE == false)&(recindex < maxindex))
    {
      switch (recordptr[recindex])
       {
         case formfeed:             /* if current byte is a form feed   */
            switch (recindex)
             {
                case first:               /* if record begins with a FF */
                   switches.shiftinp =    /* then no shift is necessary */
                                false;    /* even if the data came from */
                                          /* input on initial call.     */

                   recordptr[recindex] =  /* translate formfeed        */
                                 skip2ch1;

                   break;                 /* Continue scanning and      */

                default:                  /* FF found in middle or end  */
                   switches.moredata = true;
                   exitstruc->request
                             = INP_INSERT;/* set call back needed flag  */
                   DONE = true;           /* set flag to exit loop      */
                   break;
             }                            /* end SWITCH statement       */

            break;                  /* end case FORM FEED               */


     /***********************************************************/
     /*  If a carriage return is found, then one scenario is    */
     /*  for a carriage return at the end of the input record   */
     /*  The other is for a carriage return occurring in the    */
     /*  middle of the record.                                  */
     /***********************************************************/

         case crgret:               /* if current byte is a carriage ret*/
            switch (recindex)
             {
                case first:               /* if record begins with a CR */
                                          /* then, set byte 0.  Continue*/
                                          /* scanning for next CR or FF */
                   if (switches.fromstg   /* translate a CR occurring in*/
                                 == true) /* the first byte of examined */
                      recordptr[0] =      /* data.  If CR is from data  */
                                prntline; /* in storage, then it just   */
                                          /* return the carriage.  This */
                                          /* allows for underscoring.   */
                   else                   /* otherwise return the car-  */
                    {                     /* riage and advance one line.*/
                      recordptr[0] = space;
                    }
                   switches.shiftinp =    /* no shift is necessary even */
                                false;    /* if the data came from      */
                                          /* input on initial call.     */
                   break;

                default:                  /* CR found in middle or end  */
                                          /* of record.                 */
                   if (recindex == (maxindex - 1))/* CR found at end of rec. */
                    {                     /* delete the byte by not     */
                                          /* requesting a callback.     */
                      switches.delchar = true;
                      exitstruc->request
                             = INP_USE;   /* set no call back needed    */
                    }
                   else
                    {
                      switches.moredata = true;
                      exitstruc->request  /* CR found in middle         */
                             = INP_INSERT;/* set call back needed       */
                    }

                   DONE = true;           /* set flag to exit loop      */
                   break;
             }                            /* end SWITCH statement       */
            break;

         default:
            break;
       }                            /* end SWITCH statement             */
      recindex++;                   /* increment the input record index */
    }                               /* end WHILE loop                   */


      /***********************************************************/
      /*  Set length of data left to be examined and length of   */
      /*  data to be returned.  The length of data to be returned*/
      /*  may increase by 1 if a shift is necessary.             */
      /***********************************************************/

   if (switches.delchar == true)    /* set length of data left to exam- */
    {
      remlength = 0;                /* ine.  If deleting a character,   */
      returnlen = recordlen - 1;    /* then nothing left to examine.    */
    }                               /* delete only last characters in a */
                                    /* record.  Currently only trailing */
                                    /* carriage return.                 */
   else
      if (switches.moredata == true)/* if moredata = true then there is */
       {                            /* more data to process.  set length*/
         remlength = recordlen -    /* of data yet to be processed.     */
                      recindex + 1; /*                                  */
         returnlen = recindex - 1;
       }
      else                          /* all data was examined and no data*/
       {
         remlength = 0;             /* left to examine.                 */
         returnlen = recindex;
       }


      /***********************************************************/
      /*  Move data from storage area input area if necessary    */
      /***********************************************************/

   if (switches.fromstg == true)    /* if data to be returned is        */
    {                               /* currently in storage area, then  */
      memmove(exitstruc->record,    /* move it to input area.           */
              recordptr,
              returnlen);
    }


      /***********************************************************/
      /*  Update storage area.  Obtain storage if necessary      */
      /***********************************************************/

   if (recindex < maxindex)         /* if there remains data to process */
    {                               /* it may need to be moved to a     */
                                    /* storage area.                    */
      if (switches.fromstg == true) /* if the remaining data is already */
       {                            /* in storage, then there is no need*/
                                    /* get storage or move remaining    */
                                    /* data to storage.  Just update    */
         workarea->stgrecptr =      /* the storage record pointer.      */
               workarea->stgrecptr + returnlen;
       }
      else
       {                            /* processing data from input       */
                                    /* so obtain storage and move rest  */
                                    /* of input to storage              */
         workarea->stgbasptr =      /* get storage and set pointer to   */
                       calloc( num, /* storage.                         */
                          MAX_LEN); /*                                  */
         workarea->stgrecptr =      /* init storage record pointer      */
            workarea->stgbasptr;    /*                                  */

             /***********************************************************/
             /*  Move the remaining portion of modified input record    */
             /*  to the 32k storage area.                               */
             /***********************************************************/

         memmove(workarea->stgrecptr,
                 &recordptr[recindex-1],
                 remlength);
       }
    }
   workarea->stgreclen = remlength;/* set length of record in storage*/


      /***********************************************************/
      /*  If necessary, shift data 1 byte and set the first byte */
      /***********************************************************/

   if (switches.shiftinp == true)         /* if the lead char was not */
    {                                     /* a CR or formfeed and the */
                                          /* data is from an initial  */
                                          /* call, then it is neces-  */
                                          /* sary to insert a blank.  */
                                          /* to do this, shift the    */
                                          /* data and change the first*/
                                          /* byte to a space.         */
      memmove(&exitstruc->record[1],      /* move the data from offset*/
                 &exitstruc->record[0],   /* 0 to offset 1 for a      */
                 returnlen);              /* length of the data to be */
                                          /* returned.                */
      exitstruc->record[0] = space;       /* set the first byte       */
      returnlen++;                        /* increment record pointer */
    }

   exitstruc->recordln = returnlen;

   rc=ASC2EB(exitstruc);

      /***********************************************************/
      /*  If nothing left to examine, free any obtained storage  */
      /***********************************************************/

   if (workarea->stgbasptr != 0)          /* if there is no more data to*/
    {                                     /* be processed, then free the*/
                                          /* 32K storage area           */
      if (workarea->stgreclen == 0)
       {
         free(workarea->stgbasptr); /* storage area.                    */
         workarea->stgbasptr = 0;   /* set storage pointer to 0         */
         workarea->stgrecptr = 0;   /* set storage record pointer to 0  */
         workarea->stgreclen = 0;   /* set storage record length to 0   */
       }
    }
   return(rc);
 }

#include <iconv.h>

#define ICONV_ERROR (iconv_t) -1

long ASC2EB( INPEXIT_PARMS *exitstruc )
{
   long           rc = 0;
   size_t         bytes;
   size_t         start_index;
   char           *record;
   static iconv_t iconv_fd = ICONV_ERROR;

      /***************************************************************/
      /*           A S S U M P T I O N S ! ! !                       */
      /*                                                             */
      /* Unlike the APKA2E exit, ASC2EB assumes that the input is    */
      /* ASCII stream data, so no check is made for an AFPDS         */
      /* structured field.  Also, we assume ASCII carriage control,  */
      /* so no check is made for EBCDIC CC and TRC.                  */
      /* The CC character has already been converted as appropriate, */
      /* so conversion will start with the second byte of the record.*/
      /***************************************************************/

   if(exitstruc->eof == 'Y')
   {
      /***************************************************************/
      /* Last record has already been done, so clean up.             */
      /***************************************************************/
      if(iconv_fd != ICONV_ERROR)
         iconv_close( iconv_fd );
   }
   else
   {
      start_index = 1;
      /***************************************************************/
      /* I only want to open the converter once.                     */
      /***************************************************************/
      if(iconv_fd == ICONV_ERROR)
      {
         iconv_fd = iconv_open( "IBM-037", "IBM-850" );
         if(iconv_fd == ICONV_ERROR)
            rc = errno;
      }

      if(rc == 0)
      {
         /************************************************************/
         /* Convert the record from ASCII to EBCDIC.                 */
         /************************************************************/
         if(exitstruc->recordln > start_index)
         {
            bytes = exitstruc->recordln - start_index;
            record = exitstruc->record + start_index;

            errno = 0;
            iconv( iconv_fd,
                   &record,
                   &(bytes),
                   &record,
                   &(bytes)
                 );
            rc = errno;
         }
      }
   }

   return( rc );
}
